package com.gxzc.zen.logging.aop

import com.alibaba.dubbo.common.logger.LoggerFactory
import com.baomidou.mybatisplus.toolkit.PluginUtils
import com.gxzc.zen.logging.constants.LogConstants
import com.gxzc.zen.logging.util.MQLogUtil
import net.sf.jsqlparser.parser.CCJSqlParserUtil
import net.sf.jsqlparser.schema.Table
import net.sf.jsqlparser.statement.delete.Delete
import net.sf.jsqlparser.statement.insert.Insert
import net.sf.jsqlparser.statement.select.PlainSelect
import net.sf.jsqlparser.statement.select.Select
import net.sf.jsqlparser.statement.update.Update
import org.apache.ibatis.executor.statement.StatementHandler
import org.apache.ibatis.mapping.BoundSql
import org.apache.ibatis.mapping.MappedStatement
import org.apache.ibatis.mapping.SqlCommandType
import org.apache.ibatis.plugin.*
import org.apache.ibatis.reflection.SystemMetaObject
import org.springframework.stereotype.Component
import java.sql.Connection
import java.util.*
import java.util.concurrent.ConcurrentHashMap


/**
 * DB日志拦截器 SQLParse 拦截
 * @author NorthLan
 * @date 2018/8/24
 * @url https://noahlan.com
 */
@Intercepts(Signature(type = StatementHandler::class, method = "prepare", args = [Connection::class, Integer::class]))
@Component
open class LogDBInterceptor : Interceptor {
    companion object {
        private val logger = LoggerFactory.getLogger(LogDBInterceptor::class.java)

        // 缓存表名信息,提高性能
        private val tableNameMapping = ConcurrentHashMap<String, String>()
    }

    override fun intercept(invocation: Invocation): Any {
        logger.debug("操作数据库日志记录开始...")
        val startTime = Date()
        val statementHandler = PluginUtils.realTarget(invocation.target) as StatementHandler
        val metaObject = SystemMetaObject.forObject(statementHandler)
        // 获取SQL操作类型
        val mappedStatement = metaObject.getValue("delegate.mappedStatement") as MappedStatement
        val boundSql = metaObject.getValue("delegate.boundSql") as BoundSql
        if (boundSql.sql.startsWith("show table")) {
            return invocation.proceed()
        }

        // 消息构造
        val type: Int
        val sql = boundSql.sql


        val statement = CCJSqlParserUtil.parse(sql)

        var dbName: String? = null
        var tableName: String? = null
        val table: Table?

        // 操作类型
        when (mappedStatement.sqlCommandType) {
            SqlCommandType.SELECT -> {
                val plainSelect = (statement as Select).selectBody as PlainSelect
                table = (plainSelect.fromItem as Table)
                type = LogConstants.DB_TYPE_SELECT
            }
            SqlCommandType.DELETE -> {
                table = (statement as Delete).table
                type = LogConstants.DB_TYPE_DELETE
            }
            SqlCommandType.INSERT -> {
                table = (statement as Insert).table
                type = LogConstants.DB_TYPE_INSERT
            }
            SqlCommandType.UPDATE -> {
                table = (statement as Update).tables[0]
                type = LogConstants.DB_TYPE_UPDATE
            }
            else -> {
                // 直接跳过
                return invocation.proceed()
            }
        }

        if (table != null) {
            val tableArr = table.name.split("\\.")
            if (tableArr.size == 1) {
                tableName = tableArr[0]
            } else {
                dbName = tableArr[0]
                tableName = tableArr[1]
            }
        }
        // 执行
        val result = invocation.proceed()

        // 发消息
        MQLogUtil.logDB(type, sql, tableName, getTableNameCN(tableName, invocation.args[0] as Connection, dbName), startTime)
        return result
    }

    override fun plugin(target: Any): Any {
        return if (target is StatementHandler) {
            Plugin.wrap(target, this)
        } else {
            target
        }
    }

    override fun setProperties(properties: Properties?) {
    }

    private fun getTableNameCN(tableName: String?, conn: Connection, dbName: String? = null): String? {
        if (tableNameMapping.containsKey(tableName)) {
            return tableNameMapping[tableName]!!
        }
        // 读取表信息并存入缓存中
        val metaData = conn.metaData
        val rs = metaData.getTables(dbName, dbName, null, null)
        while (rs.next()) {
            val tName = rs.getString("TABLE_NAME")
            val remarks = rs.getString("REMARKS")
            if (tName != null && remarks != null) {
                tableNameMapping[tName] = remarks
            }
        }
        // 再次读取缓存信息
        return if (tableNameMapping.containsKey(tableName)) {
            tableNameMapping[tableName]!!
        } else {
            null
        }
    }
}